"""Common schema definitions."""
from datetime import datetime
from enum import Enum
from typing import List, Literal, Optional

from pydantic import BaseModel, Extra
from pydantic.fields import Field

CRON_REGEX = "(@(annually|yearly|monthly|weekly|daily|hourly))|" "((((\\d+,)+\\d+|(\\d+(\\/|-)\\d+)|\\d+|\\*) ?){5,7})"
DATASET_ID_REGEX = "[a-zA-Z0-9\\-_\\.]+"

DATASET_ID_DEF = Field(
    title="DatasetId",
    description="The unique ID of an dataset. This is specific to WhyLabs. If the dataset ID "
    "does not exist, user will get a validation exception when saving the "
    "config with WhyLabs API",
    regex=DATASET_ID_REGEX,
    max_length=100,
)


class NoExtrasBaseModel(BaseModel, extra=Extra.forbid):  # type: ignore
    """No extras base model.

    Inherit to prevent accidental extra fields.
    """


class ImmediateSchedule(NoExtrasBaseModel):
    """Schedule the monitor to run immediately."""

    type: Literal["immediate"] = "immediate"


class TimeRange(NoExtrasBaseModel):
    """Support for a specific time range."""

    start: datetime = Field(description="Inclusive. Start time of a time range.")
    end: datetime = Field(description="Exclusive. End time of a time range.")


class CronSchedule(NoExtrasBaseModel):
    """Support for scheduling."""

    type: Literal["cron"] = "cron"
    cron: str = Field(
        description="Cron expression",
        regex=CRON_REGEX,
    )
    exclusionRanges: Optional[List[TimeRange]] = Field(
        title="ExclusionRanges", description="The ranges of dates during which this Analyzer is NOT run."
    )
    # TODO: support other mode of configuring scheduling


class Cadence(str, Enum):
    """Cadence for an analyzer or monitor run."""

    hourly = "hourly"
    daily = "daily"
    weekly = "weekly"
    monthly = "monthly"


class FixedCadenceSchedule(NoExtrasBaseModel):
    """Support for scheduling based on a predefined cadence."""

    type: Literal["fixed"] = "fixed"
    cadence: Literal[Cadence.hourly, Cadence.daily, Cadence.weekly, Cadence.monthly] = Field(
        description="Frequency to run the analyzer or monitor, based on UTC time. The monitor will run at the start of "
        "the cadence with some SLA depending on the customer tiers.",
    )
    exclusionRanges: Optional[List[TimeRange]] = Field(
        title="ExclusionRanges", description="Ranges of dates during which this Analyzer is NOT run."
    )


class Metadata(NoExtrasBaseModel):
    """Metadata for a top-level objects such as monitors, analyzers, and schema.

    This object is managed by WhyLabs. Any user-provided values will be ignored on WhyLabs side.
    """

    version: int = Field(description="A monotonically increasing numer that indicates the version of the object.")
    schemaVersion: Optional[int] = Field(
        None,
        description="The version of the schema. Currently the accepted value is 1.",
        le=1,
        ge=1,
    )
    updatedTimestamp: int = Field(
        description="Last updated timestamp",
        gt=0,
    )
    author: str = Field(
        description="The author of the change. It can be an API Key ID, a user ID, or a WhyLabs system ID.",
        max_length=100,
        regex="[0-9a-zA-Z-_.+]+",
    )
    description: Optional[str] = Field(
        None,
        description="A description of the object",
        max_length=1000,
    )



"""Segment definitions."""
from typing import List

from pydantic import Field

from whylabs_toolkit.monitor.models.commons import NoExtrasBaseModel


class SegmentTag(NoExtrasBaseModel):
    """A single tag key value pair for a segment."""

    key: str = Field(max_length=1000)
    value: str = Field(max_length=1000)


class Segment(NoExtrasBaseModel):
    """A segment is a list of tags.

    We normalize these in the backend.
    """

    tags: List[SegmentTag] = Field(
        description="List of tags that define the specific segment",
        max_items=10,
    )



"""Schema for analyses."""
from typing import Any, Dict, List, Optional, Union

from pydantic import BaseModel, Field, constr

from whylabs_toolkit.monitor.models.commons import NoExtrasBaseModel

from ..commons import CronSchedule, FixedCadenceSchedule, Metadata
from ..utils import anyOf_to_oneOf, duration_field
from .algorithms import (
    ColumnListChangeConfig,
    ComparisonConfig,
    ListComparisonConfig,
    FrequentStringComparisonConfig,
    DiffConfig,
    DriftConfig,
    ExperimentalConfig,
    FixedThresholdsConfig,
    SeasonalConfig,
    StddevConfig,
)
from .targets import ColumnMatrix, DatasetMatrix


class Analyzer(NoExtrasBaseModel):
    """Configuration for running an analysis.

    An analysis targets a metric (note that a metric could be a complex object) for one or multiple fields in
    one or multiple segments. The output is a list of 'anomalies' that might show issues with data.
    """

    metadata: Optional[Metadata] = Field(
        None, description="WhyLabs-managed metadata. This is to track various metadata for auditing."
    )

    id: str = Field(
        None,
        description="A unique, human readable ID for an analyzer. "
        "Can only contain alpha numeric characters, underscores and dashes",
        min_length=10,
        max_length=128,
        regex="[0-9a-zA-Z\\-_]+",
    )
    displayName: Optional[str] = Field(
        None,
        id="DisplayName",
        description="A display name for the analyzer if view through WhyLabs UI. Can only contain dashes, underscores,"
        "spaces, and alphanumeric characters",
        min_length=10,
        max_length=256,
        regex="[0-9a-zA-Z \\-_]+",
    )
    tags: Optional[  # type: ignore
        List[constr(min_length=3, max_length=32, regex="[0-9a-zA-Z\\-_]")]  # noqa
    ] = Field(  # noqa F722
        None, description="A list of tags that are associated with the analyzer."
    )
    # disabling CronSchedule as it can be tricky on the BE
    schedule: Optional[FixedCadenceSchedule] = Field(  # Optional[Union[CronSchedule, FixedCadenceSchedule]] = Field(
        None,
        description="A schedule for running the analyzer. If not set, the analyzer's considered disabled",
    )
    disabled: Optional[bool] = Field(
        None,
        description="Whether the analyzer is disabled. "
        "This allows user to keep the configuration"
        "around without having to delete the analyzer config",
    )
    disableTargetRollup: Optional[bool] = Field(
        None,
        description="For customers with individual profile storage enabled on their account (contact us), this "
        "allows a user to monitor individual profiles without rolling them up. When enabled, analysis "
        "will be timestamped 1:1 with the profile's dataset timestamp rather than being truncated "
        "to the dataset granularity. ",
    )
    targetMatrix: Union[ColumnMatrix, DatasetMatrix] = Field(
        description="A matrix for possible locations of the target",
        discriminator="type",
    )
    dataReadinessDuration: Optional[str] = duration_field(
        title="DataReadinessDuration",
        description="ISO 8610 duration format. The duration determines how fast data is ready for the monitor. For "
        "example, if your pipeline takes 2 days to deliver profiles to WhyLabs, the value should be"
        "P2D. Note that this value will be used to evaluate missing data as well",
    )
    batchCoolDownPeriod: Optional[str] = duration_field(
        title="BatchCoolDownPeriod",
        description="ISO 8610 duration format. Specifies the duration that the monitor will wait from the last time"
        "a profile arrives Any batch involved in the calculation must have received the last profile by "
        "the duration.",
    )
    backfillGracePeriodDuration: Optional[str] = duration_field(
        title="BackfillGracePeriodDuration",
        description="ISO 8610 duration format. How far back an analyzer will attempt to backfill late data. Note that "
        "we will only backfill batches not previously analyzed. If the batch was already analyzed, "
        "even with partial data, the backfill will ignore the new data unless you trigger an explicit "
        "backfill request. We support 48 hours for hourly data, 30 days for daily data, and 6 months for "
        "monthly data.",
    )

    # NOT YET IMPLEMENTED:
    # ExperimentalConfig,
    # ColumnListChangeConfig,

    config: Union[
        DiffConfig,
        FixedThresholdsConfig,
        ListComparisonConfig,
        FrequentStringComparisonConfig,
        StddevConfig,
        DriftConfig,
        ComparisonConfig,
        SeasonalConfig,
    ] = Field(description="The configuration map of the analyzer", discriminator="type")

    class Config:
        """Updates JSON schema anyOf to oneOf."""

        # noinspection PyUnusedLocal
        @staticmethod
        def schema_extra(schema: Dict[str, Any], model: BaseModel) -> None:
            """Update specific fields here (for Union type, specifically)."""
            anyOf_to_oneOf(schema, "config")
            anyOf_to_oneOf(schema, "targetMatrix")



"""Schema definitions."""
from enum import Enum
from typing import Dict, List, Optional

from pydantic import Field

from whylabs_toolkit.monitor.models.commons import Metadata, NoExtrasBaseModel
from whylabs_toolkit.monitor.models.segments import Segment
from whylabs_toolkit.monitor.models.utils import COLUMN_NAME_TYPE


class ColumnDiscreteness(str, Enum):
    """Classifying the type."""

    discrete = "discrete"
    continuous = "continuous"


class ColumnDataType(str, Enum):
    """Options for configuring data type for a column."""

    integral = "integral"
    fractional = "fractional"
    boolean = "bool"
    string = "string"
    unknown = "unknown"
    null = "null"


class ColumnSchema(NoExtrasBaseModel):
    """Schema configuration for a column.

    Should be generated by WhyLabs originally but can be overridden by users.
    """

    discreteness: ColumnDiscreteness = Field(
        description="Whether a column should be discrete or continuous. WhyLabs will detect this by default but"
        "users can override this value. Changing this columns will change the default grouping (discrete"
        " columns vs. continuous columns."
    )
    dataType: ColumnDataType = Field(
        description="The data type of the columns. Setting this field affects the default grouping (i.e integral "
        "columns) and the frontend behavior (what default metrics to offer. It does NOT affect what is "
        "collected in whylogs (an integral field can still have a frequentItems sketch if it opts in)"
    )
    classifier: Optional[str] = Field(
        "input",
        description="We can classify these columns into various grouping. Currently we only support 'input' "
        "and 'output'",
        max_length=20,
    )


class WeightConfig(NoExtrasBaseModel):
    """Object that specifies column weights.

    - By default, the weight of a column is None (unspecified)
    - If the weight is unspecified, the column is EXCLUDED when you perform a filter/sort by weight
    - For sorting, unweighted column take the LEAST PRECEDENCE, meaning that weight column have higher priorities
    - They are not hierarchical: if a segment weight config is specified and a column does not have a weight in that
    config, we will not use any hierarchy to resolve the value. It will be None
    - Order of unweighted column is undefined.
    """

    weights: Dict[COLUMN_NAME_TYPE, float] = Field(description="Weights of the features")  # type: ignore


class SegmentWeightConfig(WeightConfig):
    """Object that specifies column weights for a segment."""

    segment: Optional[Segment] = Field(
        None,
        description="Set this value if the weights are for a specific segment. Note that an empty segment (no tags)"
        "is treated to be the overall segment in this case.",
    )


class EntitySchema(NoExtrasBaseModel):
    """Schema definition of an entity."""

    metadata: Optional[Metadata] = Field(
        None, description="WhyLabs-managed metadata. This is to track various metadata for auditing."
    )
    columns: Dict[COLUMN_NAME_TYPE, ColumnSchema] = Field(  # type: ignore
        description="Schema configuration for the entity"
    )


class EntityWeights(NoExtrasBaseModel):
    """Entity weight configurations."""

    metadata: Optional[Metadata] = Field(
        None, description="WhyLabs-managed metadata. This is to track various metadata for auditing."
    )
    defaultWeights: Optional[WeightConfig] = Field(
        None,
        title="DefaultWeights",
        description="Optional. Default weights for ALL the segments that don't have an explicit weight specification."
        "Note that if you specify the segment in this WeightConfig object, it is ignored.",
    )
    segmentWeights: Optional[List[SegmentWeightConfig]] = Field(
        None,
        title="SegmentWeights",
        description="Optional. Segment-specific weights. Use this if you want to override the defaultWeights. Note that"
        "there might be a case where a segment fields are weighted without specifying the default weights",
        max_items=1000,
    )



# flake8: noqa
"""Analyzer module."""
from .algorithms import *
from .analyzer import Analyzer
from .baseline import *
from .targets import *

__all__ = [
    "DatasetMetric",
    "SimpleColumnMetric",
    "ComplexMetrics",
    # analyzer
    "Analyzer",
    # baseline
    "BaselineType",
    "ReferenceProfileId",
    "TimeRangeBaseline",
    "TrailingWindowBaseline",
    "SingleBatchBaseline",
    # configs
    "DriftConfig",
    "DiffConfig",
    "ComparisonConfig",
    "ComparisonOperator",
    "ExperimentalConfig",
    "FixedThresholdsConfig",
    "ColumnListChangeConfig",
    "SeasonalConfig",
    "ListComparisonConfig",
    "FrequentStringComparisonConfig",
    "StddevConfig",
    # enums
    "DiffMode",
    "ThresholdType",
    "AlgorithmType",
    "DatasetMetric",
    "SimpleColumnMetric",
    "ComplexMetrics",
    "ListComparisonOperator",
    "FrequentStringComparisonOperator",
    # targets
    "DatasetMatrix",
    "ColumnMatrix",
    "TargetLevel",
    "ExpectedValue",
]



from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from typing import Any, Dict, List, Optional

from whylabs_client.models import EntitySchema

from whylabs_toolkit.helpers.config import Config
from whylabs_toolkit.helpers.utils import get_models_api
from whylabs_toolkit.monitor.models.column_schema import ColumnDataType

BASE_ENDPOINT = "https://api.whylabsapp.com"


@dataclass
class ColumnsClassifiers:
    inputs: List[str] = field(default_factory=list)  # type: ignore
    outputs: List[str] = field(default_factory=list)  # type: ignore


@dataclass
class ColumnsDiscreteness:
    discrete: List[str] = field(default_factory=list)  # type: ignore
    continuous: List[str] = field(default_factory=list)  # type: ignore


class UpdateEntity(ABC):
    def __init__(self, dataset_id: Optional[str] = None, org_id: Optional[str] = None, config: Config = Config()):
        self.dataset_id = dataset_id or Config().get_default_dataset_id()
        self.org_id = org_id or Config().get_default_org_id()
        self.api = get_models_api(config=config)

    def _get_entity_schema(self) -> Any:
        entity_schema = self.api.get_entity_schema(org_id=self.org_id, dataset_id=self.dataset_id)
        return entity_schema

    def _put_entity_schema(self, schema: EntitySchema) -> None:
        self.api.put_entity_schema(org_id=self.org_id, dataset_id=self.dataset_id, entity_schema=schema)

    def _get_current_entity_schema(self) -> None:
        self.current_entity_schema = self._get_entity_schema()
        self.columns_dict = self.current_entity_schema["columns"]

    @abstractmethod
    def _validate_input(self) -> None:
        pass

    @abstractmethod
    def _update_entity_schema(self) -> None:
        pass

    def _put_updated_entity_schema(self) -> None:
        metadata_dict = self.current_entity_schema["metadata"]
        entity_schema_dict = EntitySchema(columns=self.columns_dict, metadata=metadata_dict)
        self._put_entity_schema(schema=entity_schema_dict)

    def update(self) -> None:
        self._validate_input()
        self._get_current_entity_schema()
        self._update_entity_schema()
        self._put_updated_entity_schema()


class UpdateColumnClassifiers(UpdateEntity):
    def __init__(self, classifiers: ColumnsClassifiers, org_id: Optional[str] = None, dataset_id: Optional[str] = None):
        super().__init__(dataset_id, org_id)
        self.classifiers = classifiers

    def _validate_input(self) -> None:
        if self.classifiers.inputs == [] and self.classifiers.outputs == []:
            raise ValueError("You must define either input or output features to use this function.")
        same_list = [item for item in self.classifiers.inputs if item in self.classifiers.outputs]
        if same_list:
            raise ValueError(f"Column {same_list[0]} must either be input or output.")

    def _update_entity_schema(self) -> Any:
        for key in self.columns_dict.keys():
            if key in self.classifiers.inputs and self.columns_dict[key]["classifier"] != "input":
                self.columns_dict[key].classifier = "input"
            elif key in self.classifiers.outputs and self.columns_dict[key]["classifier"] != "output":
                self.columns_dict[key].classifier = "output"


class UpdateEntityDataTypes(UpdateEntity):
    """
    Update data types on each column of the dataset

    Arguments
    ----
    columns_schema: Dict[str, ColumnDataType]
        The keys are column names and the values are the
        desired data_types, as the example below shows

    ```python
    columns_schema = {
        "column_1": ColumnDataType.fractional,
        "column_2": ColumnDataType.boolean,
        "column_3": ColumnDataType.string,
    }
    ```

    These are the currently supported data types:
    ---
    - integral
    - fractional
    - bool
    - string
    - unknown
    - null
    ---
    """

    def __init__(
        self, columns_schema: Dict[str, ColumnDataType], org_id: Optional[str] = None, dataset_id: Optional[str] = None
    ):
        super().__init__(dataset_id, org_id)
        self.columns_schema = columns_schema

    def _validate_input(self) -> None:
        for data_type in self.columns_schema.values():
            if not isinstance(data_type, ColumnDataType):
                raise ValueError(
                    f"{data_type} is not an accepted data type! Refer to this functions help to learn more."
                )

    def _update_entity_schema(self) -> None:
        for column, data_type in self.columns_schema.items():
            if column in self.columns_dict.keys() and self.columns_dict[column]["data_type"] != data_type.value:
                self.columns_dict[column].data_type = self.columns_schema[column].value


class UpdateColumnsDiscreteness(UpdateEntity):
    def __init__(
        self,
        columns: ColumnsDiscreteness,
        org_id: Optional[str] = None,
        dataset_id: Optional[str] = None,
    ):
        super().__init__(dataset_id, org_id)
        self.columns = columns

    def _validate_input(self) -> None:
        if self.columns.discrete == [] and self.columns.continuous == []:
            raise ValueError("You must define either discrete or continuous columns to use this.")

        same_list = [item for item in self.columns.discrete if item in self.columns.continuous]
        if same_list:
            raise ValueError(f"Column {same_list[0]} must either be discrete or continuous.")

    def _update_entity_schema(self) -> Any:
        for key in self.columns_dict.keys():
            if key in self.columns.discrete and self.columns_dict[key]["discreteness"] != "discrete":
                self.columns_dict[key].discreteness = "discrete"
            elif key in self.columns.continuous and self.columns_dict[key]["discreteness"] != "continuous":
                self.columns_dict[key].discreteness = "continuous"



"""Define various baselines."""
from enum import Enum
from typing import List, Literal, Optional

from pydantic import Field

from whylabs_toolkit.monitor.models.commons import DATASET_ID_DEF, NoExtrasBaseModel, TimeRange


class BaselineType(str, Enum):
    """Supported baseline types."""

    BatchTimestamp = "BatchTimestamp"
    Reference = "Reference"
    TrailingWindow = "TrailingWindow"
    TimeRange = "TimeRange"
    CurrentBatch = "CurrentBatch"


class _Baseline(NoExtrasBaseModel):
    """Base class for a baseline."""

    datasetId: Optional[str] = DATASET_ID_DEF


class _SegmentBaseline(_Baseline):
    inheritSegment: Optional[bool] = Field(
        None,
        title="InheritSegment",
        description="Default to false. Whether to use the segment from the target to filter down the baseline",
    )


class ReferenceProfileId(_Baseline):
    """A baseline based on a static reference profile.

    A typical use case is to use a "gold" dataset and upload its profile to WhyLabs. This can be a training dataset
    as well for an ML model.
    """

    type: Literal[BaselineType.Reference] = BaselineType.Reference
    profileId: str = Field(
        title="ProfileId",
        description="The unique profile ID for the reference profile",
        max_length=100,
    )


class TrailingWindowBaseline(_SegmentBaseline):
    """A dynamic trailing window.

    This is useful if you don't have a static baseline to monitor against. This is the default mode for most
    monitors.
    """

    type: Optional[Literal[BaselineType.TrailingWindow]] = Field(BaselineType.TrailingWindow)
    size: int = Field(description="Window size", le=90, gt=3)
    offset: Optional[int] = Field(
        None,
        description="Offset from the current batch for the range of the trailing window. Default to 1 (the previous "
        "batch). This means that if set this to 0, the baseline will include the current batch's value, or"
        "if we set it o 7, then the window is off by 7.",
    )
    exclusionRanges: Optional[List[TimeRange]] = Field(
        None,
        title="ExclusionRanges",
        description="The list of exclusion ranges",
        max_items=100,
    )


class TimeRangeBaseline(_SegmentBaseline):
    """A static time range.

    Instead of using a single profile or a trailing window, user can lock in a "good" period.
    """

    type: Literal[BaselineType.TimeRange] = BaselineType.TimeRange
    range: TimeRange = Field(description="The range to set the time range with")


class SingleBatchBaseline(_SegmentBaseline):
    """Using current batch.

    This is used when you want to use one batch to monitor another batch in a different metric entity.
    """

    type: Literal[BaselineType.CurrentBatch] = BaselineType.CurrentBatch
    offset: Optional[int] = Field(
        None,
        description="Offset from the current batch for the baseline. Default to 0 - (the current batch). This means "
        "that if this field set this to 0, the baseline be the current batch's value. The dataset field"
        "is required to be set for this baseline config."
        "Typical use case is to use another entity to monitor against the current entity",
    )
    datasetId: str = DATASET_ID_DEF



import logging
from typing import Any, List, Optional

from whylabs_client.exceptions import ApiValueError
from whylabs_client.exceptions import NotFoundException, ForbiddenException

from whylabs_toolkit.helpers.config import Config
from whylabs_toolkit.helpers.utils import get_monitor_api, get_models_api
from whylabs_toolkit.utils.granularity import Granularity


BASE_ENDPOINT = "https://api.whylabsapp.com"
logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)


# TODO create deactivate_monitor


def get_monitor_config(org_id: str, dataset_id: str, config: Config = Config()) -> Any:
    api = get_monitor_api(config=config)
    try:
        monitor_config = api.get_monitor_config_v3(org_id=org_id, dataset_id=dataset_id)
        return monitor_config
    except NotFoundException:
        logger.warning(f"Could not find a monitor config for {dataset_id}")
        return None


def get_monitor(
    monitor_id: str, org_id: Optional[str] = None, dataset_id: Optional[str] = None, config: Config = Config()
) -> Any:
    org_id = org_id or config.get_default_org_id()
    dataset_id = dataset_id or config.get_default_dataset_id()

    api = get_monitor_api(config=config)
    try:
        monitor = api.get_monitor(org_id=org_id, dataset_id=dataset_id, monitor_id=monitor_id)
        return monitor
    except (ForbiddenException, NotFoundException):
        logger.warning(
            f"Could not find a monitor with id {monitor_id} for {dataset_id}." "Did you set a correct WHYLABS_API_KEY?"
        )
        return None


def get_analyzer_ids(
    monitor_id: str, org_id: Optional[str] = None, dataset_id: Optional[str] = None, config: Config = Config()
) -> Any:
    org_id = org_id or config.get_default_org_id()
    dataset_id = dataset_id or config.get_default_dataset_id()
    try:
        monitor_config = get_monitor_config(org_id=org_id, dataset_id=dataset_id, config=config)

        if monitor_config:
            for item in monitor_config.get("monitors"):
                if item["id"] == monitor_id:
                    resp = item["analyzerIds"]
                    return resp
    except ForbiddenException:
        logger.warning(f"Could not find analyzer IDs for {org_id}, {dataset_id}, {monitor_id}")
        return None


def get_analyzers(
    monitor_id: str, org_id: Optional[str] = None, dataset_id: Optional[str] = None, config: Config = Config()
) -> Optional[List[Any]]:
    org_id = org_id or config.get_default_org_id()
    dataset_id = dataset_id or config.get_default_dataset_id()
    api = get_monitor_api(config=config)
    analyzers = []
    analyzer_ids = get_analyzer_ids(org_id=org_id, dataset_id=dataset_id, monitor_id=monitor_id, config=config)
    if analyzer_ids:
        for analyzer in analyzer_ids:
            analyzers.append(api.get_analyzer(org_id=org_id, dataset_id=dataset_id, analyzer_id=analyzer))
        return analyzers
    else:
        return None


def get_model_granularity(
    org_id: Optional[str] = None, dataset_id: Optional[str] = None, config: Config = Config()
) -> Optional[Granularity]:
    org_id = org_id or config.get_default_org_id()
    dataset_id = dataset_id or config.get_default_dataset_id()

    api = get_models_api(config=config)
    model_meta = api.get_model(org_id=org_id, model_id=dataset_id)

    time_period_to_gran = {
        "H": Granularity.hourly,
        "D": Granularity.daily,
        "W": Granularity.weekly,
        "M": Granularity.monthly,
    }
    if model_meta:
        for key, value in time_period_to_gran.items():
            if key in model_meta["time_period"]:
                return value
    return None


def delete_monitor(
    monitor_id: str, org_id: Optional[str] = None, dataset_id: Optional[str] = None, config: Config = Config()
) -> None:
    org_id = org_id or config.get_default_org_id()
    dataset_id = dataset_id or config.get_default_dataset_id()

    api = get_monitor_api(config=config)
    try:
        analyzer_ids = get_analyzer_ids(org_id=org_id, dataset_id=dataset_id, monitor_id=monitor_id, config=config)
        if analyzer_ids is None:
            return
        for analyzer_id in analyzer_ids:
            resp_analyzer = api.delete_analyzer(org_id=org_id, dataset_id=dataset_id, analyzer_id=analyzer_id)
            logger.debug(f"Deleted analyzer with Resp:{resp_analyzer}")
        resp_monitor = api.delete_monitor(org_id=org_id, dataset_id=dataset_id, monitor_id=monitor_id)
        logger.debug(f"Deleted monitor with Resp:{resp_monitor}")
    except ApiValueError as e:
        logger.error(f"Error deleting monitor {monitor_id}: {e.msg}")
        raise e



from enum import Enum
from dataclasses import dataclass
from whylogs.core.schema import DatasetSchema
from typing import Optional


class DatasetCadence(Enum):
    HOURLY = "HOURLY"
    DAILY = "DAILY"


class DatasetUploadCadenceGranularity(Enum):
    MINUTE = "M"
    HOUR = "H"
    DAY = "D"


@dataclass
class DatasetUploadCadence:
    interval: int
    granularity: DatasetUploadCadenceGranularity


@dataclass
class DatasetOptions:
    schema: Optional[DatasetSchema]
    dataset_cadence: DatasetCadence
    whylabs_upload_cadence: DatasetUploadCadence



"""The overall document for monitor."""
import uuid
from typing import List, Literal, Optional

from pydantic import Field

from whylabs_toolkit.monitor.models.commons import NoExtrasBaseModel

from .analyzer import Analyzer
from .column_schema import EntitySchema, EntityWeights
from .commons import DATASET_ID_DEF, Metadata
from .monitor import Monitor
from ...utils.granularity import Granularity


class Document(NoExtrasBaseModel):
    """The main document that dictates how the monitor should be run. This document is managed by WhyLabs internally."""

    id: Optional[uuid.UUID] = Field(None, description="A unique ID for the document")
    schemaVersion: Literal[1] = Field(
        description="The schema version of the document",
        title="SchemaVersion",
        default=1,
    )
    metadata: Optional[Metadata] = Field(
        None, description="WhyLabs-managed metadata. This is to track various metadata for auditing."
    )
    orgId: str = Field(title="OrgId", description="Organization ID for the document", max_length=100)
    datasetId: str = DATASET_ID_DEF
    granularity: Granularity = Field(description="Granularity of the entity")
    allowPartialTargetBatches: Optional[bool] = Field(
        None,
        title="AllowPartialTargetBatches",
        description="""The standard 
        flow waits for a target batch as defined by the dataset granularity 
        setting to conclude before running analysis. For example, on monthly datasets datapoints in the 
        current month would be analyzed at midnight on the last day of the month anticipating additional 
        data may be profiled. With allowPartialTargetBatches enabled a target batch may be analyzed as 
        soon as the data is present and dataReadinessDuration/batchCooldownPeriod (if configured) 
        conditions have been met. This can be ideal for data pipelines that upload a single profile per 
        dataset granularity to reduce the waiting time for analysis.""",
    )
    entitySchema: Optional[EntitySchema] = Field(description="Schema configuration for the entity")
    weightConfig: Optional[EntityWeights] = Field(None, description="Weight configuration for the entity")
    analyzers: List[Analyzer] = Field(description="List of analyzers", max_items=1000)
    monitors: List[Monitor] = Field(description="List of monitors", max_items=1000)



"""Schema for configuring a monitor."""
from enum import Enum
from typing import Any, Dict, List, Literal, Optional, Union

from pydantic import BaseModel, Field, constr, HttpUrl, parse_obj_as, validator


from whylabs_toolkit.monitor.models.commons import (
    CronSchedule,
    FixedCadenceSchedule,
    ImmediateSchedule,
    Metadata,
    NoExtrasBaseModel,
)
from whylabs_toolkit.monitor.models.utils import COLUMN_NAME_TYPE, METRIC_NAME_STR, anyOf_to_oneOf


class MonitorConfigMetadata(NoExtrasBaseModel):
    """Metadata related to a monitor."""

    revision: int = Field(title="Revision number")
    update_timestamp: int = Field(title="Last update timestamp of this config")
    update_author: str = Field(title="The entity that updated this config", max_length=1000)


class GlobalAction(NoExtrasBaseModel):
    """Actions that are configured at the team/organization level."""

    type: Literal["global"] = "global"
    target: str = Field(description="The unique action ID in the platform", regex="[a-zA-Z0-9\\-_]+", max_length=100)


class EmailRecipient(NoExtrasBaseModel):
    """Action to send an email."""

    type: Literal["email"] = "email"
    id: str = Field(description="The e-mail ID to which you wish to send notifications to")
    destination: str = Field(description="Destination email", format="email", max_length=1000)


class SlackWebhook(NoExtrasBaseModel):
    """Action to send a Slack webhook."""

    type: Literal["slack"] = "slack"
    id: str = Field(description="The endpoint ID to which you wish to send notifications to")
    destination: str = Field(description="The Slack target webhook endpoint")


class RawWebhook(NoExtrasBaseModel):
    """Action to send a Raw webhook."""

    type: Literal["raw"] = "raw"
    id: str = Field(description="The endpoint ID to which you wish to send notifications to")
    destination: Optional[str] = Field(
        default=None, description="Sending raw unformatted message in JSON format to a webhook"
    )


class PagerDuty(NoExtrasBaseModel):
    """Action to send a PagerDuty notification."""

    type: Literal["pager_duty"] = "pager_duty"
    id: str = Field(description="The PagerDuty endpoint ID to send notifications to")
    destination: Optional[str] = Field(
        default=None,
        description="The secret key to access the PagerDuty endpoint. Required when the ID was not created",
    )


class AnomalyFilter(NoExtrasBaseModel):
    """Filter the anomalies based on certain criteria. If the alerts are filtered down to 0, the monitor won't fire."""

    includeColumns: Optional[List[COLUMN_NAME_TYPE]] = Field(  # type: ignore
        None,
        title="IncludeColumns",
        description="If set, we only include anomalies from these columns",
        max_items=1000,
    )
    excludeColumns: Optional[List[COLUMN_NAME_TYPE]] = Field(  # type: ignore
        None,
        title="ExcludeColumns",
        description="If set, we will exclude anomalies from these columns. This is applied AFTER the includeColumns",
        max_items=1000,
    )
    minWeight: Optional[float] = Field(
        None,
        title="MinWeight",
        description="We will include only features with weights greater "
        "than or equal to this value. NOT SUPPORTED YET",
    )
    maxWeight: Optional[float] = Field(
        None,
        title="MaxWeight",
        description="We will include only features with weights less than" "or equal to this value. NOT SUPPORTED YET",
    )
    minRankByWeight: Optional[int] = Field(
        None,
        title="MinRankByWeight",
        description="Include only features ranked greater than or equal to"
        "this value by weight. If features have the same weight"
        ", we order them alphabetically. NOT SUPPORTED YET",
    )
    maxRankByWeight: Optional[int] = Field(
        None,
        title="MaxRankByWeight",
        description="Include only features ranked less than or equal to"
        "this value by weight. If features have the same "
        "weight, we order them alphabetically. NOT "
        "SUPPORTED YET",
    )
    minTotalWeight: Optional[float] = Field(
        None,
        title="MinTotalWeight",
        description="Only fire the monitor if the total weights of the"
        " alerts (based on feature weights) is greater than or "
        "equal to this value. NOT SUPPORTED YET",
    )
    maxTotalWeight: Optional[float] = Field(
        None,
        title="MaxTotalWeight",
        description="Only fire the monitor if the total weights of the"
        " alerts (based on feature weights) is less than or "
        "equal to this value. NOT SUPPORTED YET",
    )
    minAlertCount: Optional[int] = Field(
        None,
        title="MinAlertCount",
        description="If the total alert count is less than this value, the " "monitor won't fire. NOT SUPPORTED YET",
    )
    maxAlertCount: Optional[int] = Field(
        None,
        title="MaxAlertCount",
        description="If the total alert count is greater than this value, " "the monitor won't fire. NOT SUPPORTED YET",
    )
    includeMetrics: Optional[List[METRIC_NAME_STR]] = Field(  # type: ignore
        None,
        title="IncludeMetrics",
        description="Metrics to filter by. NOT SUPPORTED YET",
        max_items=100,
    )


excludeMetrics: Optional[List[METRIC_NAME_STR]] = Field(  # type: ignore
    None,
    title="ExcludeMetrics",
    description="Metrics to filter by. NOT SUPPORTED YET",
    max_items=100,
)


class EveryAnomalyMode(NoExtrasBaseModel):
    """Config mode that indicates the monitor will send out individual messages per anomaly."""

    type: Literal["EVERY_ANOMALY"] = "EVERY_ANOMALY"
    filter: Optional[AnomalyFilter] = Field(None, description="Filter for anomalies")


class DigestModeGrouping(str, Enum):
    """Enable the ability to group digest by various fields."""

    byField = "byColumn"
    byDataset = "byDataset"
    byAnalyzer = "byAnalyzer"
    byDay = "byDay"
    byHour = "byHour"


class DigestMode(NoExtrasBaseModel):
    """Config mode that indicates the monitor will send out a digest message."""

    type: Literal["DIGEST"] = Field("DIGEST")
    filter: Optional[AnomalyFilter] = Field(None, description="Filter for anomalies")
    creationTimeOffset: Optional[str] = Field(
        None,
        # format='duration', # TODO: is not supported by draft-7, only in draft 2019
        title="CreationTimeOffset",
        description="Optional for Immediate digest, required for Scheduled digest. The earliest creation timestamp"
        " that we will "
        "filter by to build the digest. ISO 8601 "
        "format for timedelta.",
        max_length=20,
    )
    datasetTimestampOffset: Optional[str] = Field(
        None,
        # format='duration',
        title="DatasetTimestampOffset",
        description="Optional for Immediate digest, required for Scheduled digest. "
        "The earliest dataset timestamp that we will filter by in the digest",
        max_length=20,
    )
    groupBy: Optional[List[DigestModeGrouping]] = Field(
        None,
        description="Default is None.If this is set, we will group alerts by these groupings and emit multiple messages"
        " per group.",
        max_items=10,
    )


class Monitor(NoExtrasBaseModel):
    """Customer specified monitor configs."""

    metadata: Optional[Metadata] = Field(None, description="Meta. This is to track various metadata for auditing.")
    id: str = Field(
        None,
        description="A human-readable alias for a monitor. Must be readable",
        min_length=10,
        max_length=128,
        regex="[0-9a-zA-Z\\-_]+",
    )
    displayName: Optional[str] = Field(
        None,
        id="DisplayName",
        description="A display name for the monitor if view through WhyLabs UI. Can only contain dashes, underscores,"
        "spaces, and alphanumeric characters",
        min_length=10,
        max_length=256,
        regex="[0-9a-zA-Z \\-_]+",
    )
    tags: Optional[  # type: ignore
        List[constr(min_length=3, max_length=32, regex="[0-9a-zA-Z\\-_]")]  # noqa F722
    ] = Field(None, description="A list of tags that are associated with the monitor.")
    analyzerIds: List[constr(regex="^[A-Za-z0-9_\\-]+$")] = Field(  # type: ignore # noqa: F722
        title="AnalyzerIds",
        description="The corresponding analyzer ID. Even though it's plural, we only support one analyzer at the "
        "moment",
        max_items=100,
    )
    schedule: Union[FixedCadenceSchedule, CronSchedule, ImmediateSchedule] = Field(
        description="Schedule of the monitor. We only support hourly monitor at " "the finest granularity",
    )
    disabled: Optional[bool] = Field(None, description="Whether the monitor is enabled or not")
    severity: Optional[int] = Field(3, description="The severity of the monitor messages")
    mode: Union[EveryAnomalyMode, DigestMode] = Field(
        description="Notification mode and how we might handle different analysis",
        discriminator="type",
    )
    actions: List[Union[GlobalAction, EmailRecipient, SlackWebhook, PagerDuty]] = Field(
        description="List of destination for the outgoing messages",
        max_items=100,
    )

    class Config:
        """Updates JSON schema anyOf to oneOf."""

        # noinspection PyUnusedLocal
        @staticmethod
        def schema_extra(schema: Dict[str, Any], model: BaseModel) -> None:
            """Update specific fields here (for Union type, specifically)."""
            anyOf_to_oneOf(schema, "mode")
            anyOf_to_oneOf(schema, "schedule")



import re
import logging
from datetime import datetime, timezone
from typing import Optional, List, Union, Any

from whylabs_client.exceptions import NotFoundException

from whylabs_toolkit.helpers.utils import get_models_api
from whylabs_toolkit.monitor.models import *
from whylabs_toolkit.monitor.models.analyzer.targets import ColumnGroups
from whylabs_toolkit.monitor.manager.credentials import MonitorCredentials
from whylabs_toolkit.helpers.monitor_helpers import get_analyzers, get_monitor, get_model_granularity
from whylabs_toolkit.helpers.config import Config


logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)

TAG_ANALYZER_CONSTRAINT = "whylabs.constraint"


class MonitorSetup:
    def __init__(self, monitor_id: str, dataset_id: Optional[str] = None, config: Config = Config()) -> None:

        self.credentials = MonitorCredentials(monitor_id=monitor_id, dataset_id=dataset_id, config=config)
        self._config = config
        self.monitor: Optional[Monitor] = self._check_if_monitor_exists()
        self.analyzer: Optional[Analyzer] = self._check_if_analyzer_exists()

        self._models_api = get_models_api(config=self._config)

        self._monitor_mode: Optional[Union[EveryAnomalyMode, DigestMode]] = None
        self._monitor_actions: Optional[List[Union[GlobalAction, EmailRecipient, SlackWebhook, PagerDuty]]] = None
        self._analyzer_schedule: Optional[FixedCadenceSchedule] = None
        self._target_matrix: Optional[Union[ColumnMatrix, DatasetMatrix]] = None
        self._analyzer_config: Optional[
            Union[
                DiffConfig,
                FixedThresholdsConfig,
                StddevConfig,
                DriftConfig,
                ComparisonConfig,
                SeasonalConfig,
                FrequentStringComparisonConfig,
                ListComparisonConfig,
            ]
        ] = None
        self._target_columns: Optional[List[str]] = []
        self._exclude_columns: Optional[List[str]] = []
        self._monitor_tags: Optional[List[str]] = []
        self._analyzer_tags: Optional[List[str]] = []
        self._analyzer_disable_target_rollup: Optional[bool] = None
        self._data_readiness_duration: Optional[str] = None

        self._prefill_properties()

    def _check_if_monitor_exists(self) -> Any:
        existing_monitor = get_monitor(
            org_id=self.credentials.org_id,
            dataset_id=self.credentials.dataset_id,
            monitor_id=self.credentials.monitor_id,
            config=self._config,
        )
        if existing_monitor:
            existing_monitor = Monitor.parse_obj(existing_monitor)
            logger.info(f"Got existing {self.credentials.monitor_id} from WhyLabs!")
        else:
            logger.info(f"Did not find a monitor with {self.credentials.monitor_id}, creating a new one.")
            existing_monitor = None
        return existing_monitor

    def _check_if_analyzer_exists(self) -> Any:
        existing_analyzers = get_analyzers(
            org_id=self.credentials.org_id,
            dataset_id=self.credentials.dataset_id,
            monitor_id=self.credentials.monitor_id,
            config=self._config,
        )
        if existing_analyzers:
            existing_analyzer = Analyzer.parse_obj(existing_analyzers[0])  # enforcing 1:1 relationship

        else:
            existing_analyzer = None
        return existing_analyzer

    def _prefill_properties(self) -> None:
        if self.monitor:
            self._monitor_mode = self.monitor.mode
            self._monitor_actions = self.monitor.actions
            self._monitor_tags = self.monitor.tags
        if self.analyzer:
            self._analyzer_schedule = self.analyzer.schedule
            self._target_matrix = self.analyzer.targetMatrix
            self._analyzer_config = self.analyzer.config
            self._analyzer_tags = self.analyzer.tags
            self._analyzer_disable_target_rollup = self.analyzer.disableTargetRollup

    @property
    def schedule(self) -> Optional[Union[CronSchedule, FixedCadenceSchedule]]:
        return self._analyzer_schedule

    @schedule.setter
    def schedule(self, schedule: FixedCadenceSchedule) -> None:
        self._analyzer_schedule = schedule

    @property
    def target_matrix(self) -> Optional[Union[ColumnMatrix, DatasetMatrix]]:
        return self._target_matrix

    @target_matrix.setter
    def target_matrix(self, target: Union[ColumnMatrix, DatasetMatrix]) -> None:
        self._target_matrix = target

    @property
    def config(
        self,
    ) -> Optional[
        Union[
            DiffConfig,
            FixedThresholdsConfig,
            StddevConfig,
            DriftConfig,
            ComparisonConfig,
            SeasonalConfig,
            FrequentStringComparisonConfig,
            ListComparisonConfig,
        ]
    ]:
        return self._analyzer_config

    @config.setter
    def config(
        self,
        config: Union[
            DiffConfig,
            FixedThresholdsConfig,
            StddevConfig,
            DriftConfig,
            SeasonalConfig,
            FrequentStringComparisonConfig,
            ListComparisonConfig,
        ],
    ) -> None:
        self._analyzer_config = config

    @property
    def actions(self) -> Optional[List[Union[GlobalAction, EmailRecipient, SlackWebhook, PagerDuty]]]:
        return self._monitor_actions

    @actions.setter
    def actions(self, actions: List[Union[GlobalAction, EmailRecipient, SlackWebhook, PagerDuty]]) -> None:
        self._monitor_actions = actions

    @property
    def mode(self) -> Optional[Union[EveryAnomalyMode, DigestMode]]:
        return self._monitor_mode

    @mode.setter
    def mode(self, mode: Union[EveryAnomalyMode, DigestMode]) -> None:
        self._monitor_mode = mode

    @property
    def is_constraint(self) -> Optional[bool]:
        return self._analyzer_tags is not None and TAG_ANALYZER_CONSTRAINT in self._analyzer_tags

    @is_constraint.setter
    def is_constraint(self, is_constraint: bool) -> None:
        if self._analyzer_config is None:
            raise ValueError("Config must first be set")
        tags = set(self._analyzer_tags or [])
        if is_constraint:
            if not isinstance(self._analyzer_config, (FixedThresholdsConfig)):
                raise ValueError("Constraint can only be set with FixedThresholdsConfig")
            tags.add(TAG_ANALYZER_CONSTRAINT)
        else:
            tags.discard(TAG_ANALYZER_CONSTRAINT)
        self._analyzer_tags = list(tags)

    @property
    def disable_target_rollup(self) -> Optional[bool]:
        return self._analyzer_disable_target_rollup == True

    @disable_target_rollup.setter
    def disable_target_rollup(self, disable_target_rollup: bool) -> None:
        self._analyzer_disable_target_rollup = disable_target_rollup

    @property
    def data_readiness_duration(self) -> Optional[str]:
        return self._data_readiness_duration

    @data_readiness_duration.setter
    def data_readiness_duration(self, delay_duration: str) -> None:
        if self._validate_delay_duration(delay_duration) is False:
            raise ValueError(f"{delay_duration} does not respect ISO 8601 format")
        self._data_readiness_duration = delay_duration

    def _validate_delay_duration(self, delay: str) -> bool:
        pattern = r"^P(\d+Y)?(\d+M)?(\d+D)?(T(\d+H)?(\d+M)?(\d+(\.\d+)?S)?)?$"
        return bool(re.match(pattern, delay))

    def _validate_columns_input(self, columns: List[str]) -> bool:
        if type(columns) != list or not all(isinstance(column, str) for column in columns):
            raise ValueError("columns argument must be a List of strings")

        group_columns = ColumnGroups.__members__.values()
        allowed_groups = any(col in group_columns for col in columns)

        if allowed_groups:
            return True

        if "group:" in columns and not allowed_groups:
            raise ValueError(f"group:[group_type] should be one of {group_columns}")

        schema = self._models_api.get_entity_schema(
            org_id=self.credentials.org_id, dataset_id=self.credentials.dataset_id
        )
        columns_dict = schema["columns"]

        for col in columns:
            if col not in columns_dict.keys():
                raise ValueError(
                    f"{col} is not present on {self.credentials.dataset_id}. Available columns are: {columns_dict.keys()}"
                )

        return True

    def set_target_columns(self, columns: List[str]) -> None:
        """
        Args:
            :columns: A list of the targeted columns to monitor against the reference profile.
            :type columns: List[str]
        """
        if self._validate_columns_input(columns=columns):
            self._target_columns = columns
            self._target_matrix = self._target_matrix or ColumnMatrix(
                include=self._target_columns, exclude=self._exclude_columns, segments=[]
            )
            if isinstance(self._target_matrix, ColumnMatrix):
                self._target_matrix.include = self._target_columns

    def exclude_target_columns(self, columns: List[str]) -> None:
        if self._validate_columns_input(columns=columns):
            self._exclude_columns = columns
            self._target_matrix = self._target_matrix or ColumnMatrix(
                include=self._target_columns, exclude=self._exclude_columns, segments=[]
            )
            if isinstance(self._target_matrix, ColumnMatrix):
                self._target_matrix.exclude = self._exclude_columns

    def set_fixed_dates_baseline(self, start_date: datetime, end_date: datetime) -> None:
        if not start_date.tzinfo:
            start_date = start_date.replace(tzinfo=timezone.utc)
        if not end_date.tzinfo:
            end_date = end_date.replace(tzinfo=timezone.utc)
        # TODO make baseline nullable and default baseline to be TrailingWindowBaseline(size=14)
        self._analyzer_config.baseline = TimeRangeBaseline(  # type: ignore
            range=TimeRange(start=start_date, end=end_date)
        )

    def __set_analyzer(self) -> None:
        self.__configure_target_matrix()

        self.__set_dataset_matrix_for_dataset_metric()
        self.__set_dataset_matrix_for_missing_data_metric()

        self.analyzer = Analyzer(
            id=self.credentials.analyzer_id,
            displayName=self.credentials.analyzer_id,
            disableTargetRollup=self._analyzer_disable_target_rollup,
            targetMatrix=self._target_matrix,
            dataReadinessDuration=self._data_readiness_duration,
            tags=self._analyzer_tags,
            schedule=self._analyzer_schedule,
            config=self._analyzer_config,
        )

    def __set_monitor(
        self, monitor_mode: Optional[Union[EveryAnomalyMode, DigestMode]], monitor_actions: Optional[List[Any]]
    ) -> None:
        self.monitor = Monitor(
            id=self.credentials.monitor_id,
            disabled=False,
            displayName=self.credentials.monitor_id,
            tags=self._monitor_tags,
            analyzerIds=[self.credentials.analyzer_id],
            schedule=ImmediateSchedule(),
            mode=monitor_mode,
            actions=monitor_actions,
        )

    def __configure_target_matrix(self) -> None:
        self._target_matrix = self._target_matrix or ColumnMatrix(
            include=self._target_columns or ["*"], exclude=self._exclude_columns, segments=[]
        )

    def __set_dataset_matrix_for_dataset_metric(self) -> None:
        if self._analyzer_config:
            if isinstance(self._analyzer_config.metric, DatasetMetric) and isinstance(
                self._target_matrix, ColumnMatrix
            ):
                logger.warning(
                    "ColumnMatrix is not configurable with a DatasetMetric." "Changing it to DatasetMatrix instead"
                )
                self._target_matrix = DatasetMatrix(segments=self._target_matrix.segments)
                return None

            elif isinstance(self._target_matrix, DatasetMatrix) and not isinstance(
                self._analyzer_config.metric, DatasetMetric
            ):
                logger.warning(
                    "Setting a DatasetMatrix requires a DatasetMetric to be used"
                    "Changing it to an empty ColumnMatrix instead"
                )
                self._target_matrix = ColumnMatrix(
                    include=self._target_columns or ["*"],
                    exclude=self._exclude_columns,
                    segments=self._target_matrix.segments,
                )
                return None

    def __set_dataset_matrix_for_missing_data_metric(self) -> None:
        if (
            isinstance(self._analyzer_config, FixedThresholdsConfig)
            and self._analyzer_config.metric == "missingDataPoint"
            and isinstance(self._target_matrix, ColumnMatrix)
        ):

            logger.warning(
                "Missing data point needs to be set with target_matrix of type DatasetMatrix"
                "Changing to DatasetMatrix now."
            )
            self._target_matrix = DatasetMatrix(segments=self._target_matrix.segments)
            return None

        if (
            isinstance(self._analyzer_config, FixedThresholdsConfig)
            and self._analyzer_config.metric == "secondsSinceLastUpload"
            and isinstance(self._target_matrix, ColumnMatrix)
        ):
            logger.warning(
                "secondsSinceLastUpload needs to be set with target_matrix of type DatasetMatrix"
                "Changing to DatasetMatrix now."
            )
            self._target_matrix = DatasetMatrix(segments=self._target_matrix.segments)
            return None

    def apply(self) -> None:
        monitor_mode = self._monitor_mode or DigestMode()
        actions = self._monitor_actions or []
        self._analyzer_schedule = self._analyzer_schedule or FixedCadenceSchedule(
            cadence=get_model_granularity(
                org_id=self.credentials.org_id, dataset_id=self.credentials.dataset_id  # type: ignore
            )
        )

        self.__set_monitor(monitor_mode=monitor_mode, monitor_actions=actions)

        self.__set_analyzer()



"""Collections of support algorithms."""
from enum import Enum
from typing import Any, Dict, List, Literal, Optional, Union

from pydantic import BaseModel, Field, constr

from whylabs_toolkit.monitor.models.analyzer.baseline import (
    ReferenceProfileId,
    SingleBatchBaseline,
    TimeRangeBaseline,
    TrailingWindowBaseline,
)
from whylabs_toolkit.monitor.models.commons import NoExtrasBaseModel, TimeRange
from whylabs_toolkit.monitor.models.utils import COLUMN_NAME_TYPE, anyOf_to_oneOf


class AlgorithmType(str, Enum):
    """Specify the algorithm type."""

    expected = "expected"
    column_list = "column_list"
    comparison = "comparison"
    list_comparison = "list_comparison"
    frequent_string_comparison = "frequent_string_comparison"
    diff = "diff"
    drift = "drift"
    stddev = "stddev"
    seasonal = "seasonal"
    fixed = "fixed"
    experimental = "experimental"


class DatasetMetric(str, Enum):
    """Metrics that are applicable at the dataset level."""

    # ingestion health. null value if not ingested yet
    profile_count = "profile.count"
    profile_last_ingestion_time = "profile.last_ingestion_time"
    profile_first_ingestion_time = "profile.first_ingestion_time"

    # within the batch
    column_row_count_sum = "column_row_count_sum"
    # shape metrics?
    shape_column_count = "shape_column_count"
    shape_row_count = "shape_row_count"
    input_count = "input.count"
    output_count = "output.count"

    # classification metrics
    classification_f1 = "classification.f1"
    classification_precision = "classification.precision"
    classification_recall = "classification.recall"
    classification_accuracy = "classification.accuracy"
    classification_auc = "classification.auc"

    # regression metrics
    regression_mse = "regression.mse"
    regression_mae = "regression.mae"
    regression_rmse = "regression.rmse"

    # other metrics
    missing_data_point = "missingDatapoint"


class SimpleColumnMetric(str, Enum):
    """Simple column metrics that are basically just a single number."""

    count = "count"  # type: ignore
    median = "median"
    max = "max"
    min = "min"
    mean = "mean"
    stddev = "stddev"
    variance = "variance"
    unique_upper = "unique_upper"
    unique_upper_ratio = "unique_upper_ratio"
    unique_est = "unique_est"
    unique_est_ratio = "unique_est_ratio"
    unique_lower = "unique_lower"
    unique_lower_ratio = "unique_lower_ratio"

    # data type counts and ratios
    count_bool = "count_bool"
    count_bool_ratio = "count_bool_ratio"
    count_integral = "count_integral"
    count_integral_ratio = "count_integral_ratio"
    count_fractional = "count_fractional"
    count_fractional_ratio = "count_fractional_ratio"
    count_string = "count_string"
    count_string_ratio = "count_string_ratio"

    # also missing values
    count_null = "count_null"
    count_null_ratio = "count_null_ratio"

    # this is a string metric
    inferred_data_type = "inferred_data_type"

    # quantiles
    quantile_p5 = "quantile_5"
    quantile_p75 = "quantile_75"
    quantile_p25 = "quantile_25"
    quantile_p90 = "quantile_90"
    quantile_p95 = "quantile_95"
    quantile_p99 = "quantile_99"


class ComplexMetrics(str, Enum):
    """Sketch-based metrics that can only be processed by certain algorithms."""

    histogram = "histogram"
    frequent_items = "frequent_items"
    unique_sketch = "unique_sketch"
    # list of columns
    column_list = "column_list"


class AlgorithmConfig(NoExtrasBaseModel):
    """Base algorithm config."""

    schemaVersion: Optional[int] = Field(
        None,
        description="The schema version of an algorithm. Typically this value" "is not required.",
        title="SchemaVersion",
    )
    params: Optional[Dict[constr(max_length=100), constr(max_length=1000)]] = Field(  # type: ignore
        None,
        description="Extra parameters for the algorithm",
    )
    metric: Union[DatasetMetric, SimpleColumnMetric, constr(max_length=100)] = Field(  # type: ignore
        description="The target metric. This field cannot be change once the analyzer is created.",
    )

    class Config:
        """Updates JSON schema anyOf to oneOf for baseline."""

        # noinspection PyUnusedLocal
        @staticmethod
        def schema_extra(schema: Dict[str, Any], model: BaseModel) -> None:
            """Update specific fields here (for Union type, specifically)."""
            anyOf_to_oneOf(schema, "baseline")


class ExpectedValue(NoExtrasBaseModel):
    """Expected value: one of these fields must be set."""

    str: Optional[constr(max_length=100)]  # type: ignore
    int: Optional[int]
    float: Optional[float]


class ComparisonOperator(str, Enum):
    """Operators for performing a comparison."""

    eq = "eq"
    # Not Yet Implemented:
    # gt = "gt"
    # lt = "lt"
    # ge = "ge"
    # le = "le"


class ListComparisonOperator(str, Enum):
    """Operators for performing a comparison."""

    in_list = "in"
    not_in_list = "not_in"


class FrequentStringComparisonOperator(str, Enum):
    """Operators for performing a comparison."""

    eq = "eq"
    target_includes_all_baseline = "target_includes_all_baseline"
    baseline_includes_all_target = "baseline_includes_all_target"


class ComparisonConfig(AlgorithmConfig):
    """Compare whether the target against a value or against a baseline's metric.

    This is useful to detect data type change, for instance.
    """

    type: Literal[AlgorithmType.comparison] = AlgorithmType.comparison
    operator: ComparisonOperator = Field(
        description="The operator for the comparison. The right hand side is the target batch's metric. The left hand"
        "side is the expected value or a baseline's metric."
    )
    expected: Optional[ExpectedValue] = Field(
        None,
        description="The expected value of the equality. If the value is not set we will extract the corresponding "
        "metric from the baseline and perform the comparison",
    )
    baseline: Optional[Union[TrailingWindowBaseline, ReferenceProfileId, TimeRangeBaseline, SingleBatchBaseline]]


class ColumnListChangeConfig(AlgorithmConfig):
    """Compare whether the target is equal to a value or not.

    This is useful to detect data type change, for instance.
    """

    type: Literal[AlgorithmType.column_list] = AlgorithmType.column_list
    mode: Literal["ON_ADD_AND_REMOVE", "ON_ADD", "ON_REMOVE"] = "ON_ADD_AND_REMOVE"
    metric: Literal[ComplexMetrics.column_list]
    exclude: Optional[List[COLUMN_NAME_TYPE]] = Field(  # type: ignore
        None,
        description="Ignore these column names. User can specify a list of regex",
        max_items=1000,
    )

    baseline: Union[TrailingWindowBaseline, ReferenceProfileId, TimeRangeBaseline, SingleBatchBaseline]


class FixedThresholdsConfig(AlgorithmConfig):
    """Fixed threshold analysis.

    If user fails to set both upper bound and lower bound, this algorithm becomes a no-op.
    WhyLabs might enforce the present of either fields in the future.
    """

    type: Literal[AlgorithmType.fixed] = AlgorithmType.fixed
    upper: Optional[float] = Field(None, description="Upper bound of the static threshold")
    lower: Optional[float] = Field(None, description="Lower bound of the static threshold")


class _ThresholdBaseConfig(AlgorithmConfig):
    maxUpperThreshold: Optional[float] = Field(
        None,
        description="Capping the threshold by this value. This value only becomes effective if the calculated upper "
        "threshold from the calculation is greater than this value.",
    )

    minLowerThreshold: Optional[float] = Field(
        None,
        description="Capping the minimum threshold by this value. This value only becomes effective if the calculated "
        "lower threshold from the calculation is lesser than this value",
    )


class StddevConfig(_ThresholdBaseConfig):
    """Calculates upper bounds and lower bounds based on stddev from a series of numbers.

    An analyzer using stddev for a window of time range.

    This calculation will fall back to Poisson distribution if there is only 1 value in the baseline.
    For 2 values, we use the formula sqrt((x_i - avg(x))^2 / n - 1)
    """

    type: Optional[Literal[AlgorithmType.stddev]] = Field(AlgorithmType.stddev)
    factor: Optional[float] = Field(
        3.0, description="The multiplier used with stddev to build the upper and lower bounds."
    )
    minBatchSize: Optional[int] = Field(
        1, title="MinBatchSize", ge=1, description="Minimum number of batches that is required"
    )
    baseline: Union[TrailingWindowBaseline, TimeRangeBaseline, ReferenceProfileId]


class SeasonalConfig(_ThresholdBaseConfig):
    """An analyzer using stddev for a window of time range.

    This will fall back to Poisson distribution if there is only 1 value in the baseline.

    This only works with TrailingWindow baseline (TODO: add backend validation)
    """

    type: Literal[AlgorithmType.seasonal] = AlgorithmType.seasonal
    algorithm: Literal["arima"] = Field("arima", description="The algorithm implementation for seasonal analysis")
    minBatchSize: Optional[int] = Field(
        30,
        title="MinBatchSize",
        description="Minimum number of batches that is required",
    )
    alpha: Optional[float] = Field(
        default=0.05,
        description="significance level for the confidence interval produced around predictions. If 0.05 then the "
        "algorithm will calculate a 95% confidence interval around predictions",
    )
    baseline: TrailingWindowBaseline
    stddevTimeRanges: Optional[List[TimeRange]] = Field(
        title="StddevTimeRanges",
        description="Ranges of time where we will apply standard deviation for confidence "
        "intervals rather than the confidence interval from the algorithm. This "
        "is to prevent data from special"
        "events from making the bands very wide for timeseries-based predictions.",
    )
    stddevMaxBatchSize: Optional[int] = Field(
        description="Maxinum number of data points to consider for calculating stddev. These are the data points"
        "preceeding the target batch."
    )
    stddevFactor: Optional[float] = Field(
        default=1.0,
        description="The multiplier factor for calculating upper bounds and lower bounds from the prediction.",
    )


class DriftConfig(AlgorithmConfig):
    """An analyzer using stddev for a window of time range.

    This analysis will detect whether the data drifts or not. By default, we use hellinger distance with a threshold
    of 0.7.
    """

    type: Literal[AlgorithmType.drift] = AlgorithmType.drift
    algorithm: Literal["hellinger", "jensenshannon", "kl_divergence", "psi"] = Field(
        "hellinger", description="The algorithm to use when calculating drift."
    )
    metric: Literal[ComplexMetrics.histogram, ComplexMetrics.frequent_items]
    threshold: float = Field(
        0.7,
        description="The threshold for the distance algorithm. Depending on the algorithm, this threshold"
        "is used for greater than or less than comparison.",
    )
    minBatchSize: Optional[int] = Field(
        1,
        title="MinBatchSize",
        description="Minimum number of batches that is required",
        ge=1,
    )
    baseline: Union[TrailingWindowBaseline, ReferenceProfileId, TimeRangeBaseline, SingleBatchBaseline]


class ExperimentalConfig(AlgorithmConfig):
    """Experimental algorithm that is not standardized by the above ones yet."""

    type: Literal[AlgorithmType.experimental] = AlgorithmType.experimental
    implementation: str = Field(description="The implementation of an experimental config", max_length=100)
    baseline: Union[TrailingWindowBaseline, ReferenceProfileId, TimeRangeBaseline, SingleBatchBaseline]
    stub: Optional[AlgorithmType] = Field(description="Stub field to flow algoirthm type into the schema. Do not use.")


class ListComparisonConfig(AlgorithmConfig):
    """Compare a target list of values against a baseline list of values."""

    type: Literal[AlgorithmType.list_comparison] = AlgorithmType.list_comparison
    operator: ListComparisonOperator = Field(
        description="The operator for the comparison. The right hand side is the target batch's metric. The left hand"
        "side is the expected value or a baseline's metric."
    )
    expected: Optional[List[ExpectedValue]] = Field(
        None,
        description="The expected values of the equality. If the value is not set we will extract the corresponding "
        "metric from the baseline and perform the comparison",
    )
    baseline: Optional[Union[TrailingWindowBaseline, ReferenceProfileId, TimeRangeBaseline, SingleBatchBaseline]]


class FrequentStringComparisonConfig(AlgorithmConfig):
    """Compare whether target against a list of values."""

    type: Literal[AlgorithmType.frequent_string_comparison] = AlgorithmType.frequent_string_comparison
    metric: Literal[ComplexMetrics.frequent_items] = ComplexMetrics.frequent_items
    operator: FrequentStringComparisonOperator = Field(description="The operator for the comparison.")
    baseline: Union[TrailingWindowBaseline, ReferenceProfileId, TimeRangeBaseline, SingleBatchBaseline]


class DiffMode(str, Enum):
    """Whether to use the absolute difference or the percentage to calculate the difference."""

    abs = "abs"
    pct = "pct"


class ThresholdType(str, Enum):
    """By default, an anomaly will be generated when the target is above or below the baseline
    by the specified threshold.

    If its only desirable to alert when the target is above the
    baseline and not the other way around, specify upper for your ThresholdType."""

    lower = "lower"  # type: ignore
    upper = "upper"  # type: ignore


class DiffConfig(AlgorithmConfig):
    """Detecting the differences between two numerical metrics."""

    type: Literal[AlgorithmType.diff] = AlgorithmType.diff
    mode: DiffMode
    thresholdType: Optional[ThresholdType]
    threshold: float = Field(
        description="The minimum threshold that will trigger an anomaly. The monitor detect the difference between"
        "the target's metric and the baseline metric. Both of these metrics MUST be in rolled up form",
    )
    baseline: Union[TrailingWindowBaseline, ReferenceProfileId, TimeRangeBaseline, SingleBatchBaseline]



"""Define what targets for the analyses."""
from enum import Enum
from typing import List, Literal, Optional, Union

from pydantic import Field

from whylabs_toolkit.monitor.models.commons import NoExtrasBaseModel
from whylabs_toolkit.monitor.models.segments import Segment
from whylabs_toolkit.monitor.models.utils import COLUMN_NAME_TYPE


class TargetLevel(str, Enum):
    """Which nested level we are targeting."""

    dataset = "dataset"
    column = "column"


class _BaseMatrix(NoExtrasBaseModel):
    segments: Optional[List[Segment]] = Field(
        None,
        description="List of targeted segments. If not set, default to the overall segment",
        max_items=1000,
    )


class DatasetMatrix(_BaseMatrix):
    """Define the matrix of fields and segments to fan out for monitoring.

    .
    """

    type: Literal[TargetLevel.dataset] = Field(
        TargetLevel.dataset,
        description="Must be 'dataset' level",
    )


class ColumnGroups(str, Enum):
    """Standard column groupings."""

    group_continuous = "group:continuous"
    group_discrete = "group:discrete"

    # based on classification
    group_input = "group:input"
    group_output = "group:output"

    # based on data types
    group_bool = "group:bool"
    group_int = "group:int"
    group_frac = "group:frac"
    group_str = "group:str"


class ColumnMatrix(_BaseMatrix):
    """Define the matrix of columns and segments to fan out for monitoring."""

    type: Literal[TargetLevel.column] = TargetLevel.column
    include: Optional[List[Union[ColumnGroups, COLUMN_NAME_TYPE]]] = Field(  # type: ignore
        None,
        description="List of allowed fields/features/columns. Could be a grouping as well.",
        max_items=1000,
    )
    exclude: Optional[List[Union[ColumnGroups, COLUMN_NAME_TYPE]]] = Field(  # type: ignore
        None,
        description="List of blocked fields/features/columns. Could be a grouping as well. This setting is "
        "evaluated AFTER the 'include' field and thus should be used with caution.",
        max_items=1000,
    )
    profileId: Optional[str] = Field(
        default=None,
        description="The unique profile ID for the reference profile",
        max_length=100,
    )



import logging
from typing import Optional

from whylabs_client.exceptions import ApiValueError, ApiException
from whylabs_client.model.model_type import ModelType
from whylabs_client.model.time_period import TimePeriod
from whylabs_client.model.metric_schema import MetricSchema

from whylabs_toolkit.helpers.utils import get_models_api
from whylabs_toolkit.helpers.config import Config

logger = logging.getLogger(__name__)


def update_model_metadata(
    dataset_id: Optional[str] = None,
    org_id: Optional[str] = None,
    time_period: Optional[str] = None,
    model_type: Optional[str] = None,
    config: Config = Config(),
) -> None:
    """
    Update model attributes like model type and period.
    """
    org_id = org_id or config.get_default_org_id()
    dataset_id = dataset_id or config.get_default_dataset_id()

    api = get_models_api(config=config)

    model_metadata = api.get_model(org_id=org_id, model_id=dataset_id)
    logger.debug(f"Updating dataset with current metadata: \n {model_metadata}")

    try:
        resp = api.update_model(
            org_id=org_id,
            model_id=dataset_id,
            model_name=model_metadata["name"],
            time_period=TimePeriod(time_period) if time_period else model_metadata["time_period"],
            model_type=ModelType(model_type) if model_type else model_metadata["model_type"],
        )
        logger.debug(f"Updated sucessfully! Resp: {resp}")
    except ApiValueError as e:
        raise e


def add_custom_metric(
    label: str,
    column: str,
    default_metric: str,
    org_id: Optional[str] = None,
    dataset_id: Optional[str] = None,
    config: Config = Config(),
) -> None:

    org_id = org_id or config.get_default_org_id()
    dataset_id = dataset_id or config.get_default_dataset_id()

    api = get_models_api(config=config)
    metric_schema = MetricSchema(label=label, column=column, default_metric=default_metric)

    try:
        api.put_entity_schema_metric(org_id, dataset_id, metric_schema)
        logger.info(f"Updated entity schema metric!")
    except ApiException as e:
        logger.error("Exception when calling ModelsApi -> put_entity_schema_metric\n")
        raise e



from whylabs_client.api.dataset_profile_api import DatasetProfileApi
from whylabs_client.api.models_api import ModelsApi
from whylabs_client.api.notification_settings_api import NotificationSettingsApi
from whylabs_client.api.monitor_api import MonitorApi

from whylabs_toolkit.helpers.client import create_client
from whylabs_toolkit.helpers.config import Config


def get_models_api(config: Config = Config()) -> ModelsApi:
    return ModelsApi(api_client=create_client(config=config))


def get_dataset_profile_api(config: Config = Config()) -> DatasetProfileApi:
    return DatasetProfileApi(api_client=create_client(config=config))


def get_notification_api(config: Config = Config()) -> NotificationSettingsApi:
    return NotificationSettingsApi(api_client=create_client(config=config))


def get_monitor_api(config: Config = Config()) -> MonitorApi:
    return MonitorApi(api_client=create_client(config=config))



"""Common utilities."""
from typing import Any, Dict

from pydantic import Field, constr


def anyOf_to_oneOf(schema: Dict[str, Any], field_name: str) -> None:
    """Turn anyOf in JSON schema to oneOf.

    onfOf is much stricter and pyDantic doesn't produce this tag. We hijack the JSON schema object to
    set this correctly.

    See: https://github.com/samuelcolvin/pydantic/issues/656
    """
    cfg = schema["properties"].get(field_name)
    if cfg is None:
        return
    if cfg.get("anyOf") is None:
        return
    cfg["oneOf"] = cfg["anyOf"]
    del cfg["anyOf"]


COLUMN_NAME_TYPE = constr(max_length=1000)
METRIC_NAME_STR = constr(max_length=50)


def duration_field(description: str, title: str) -> Any:
    """Duration of a field."""
    return Field(
        None,
        title=title,
        description=description,
        example="PT1H, P1D",
        regex="^P(?!$)(\\d+M)?(\\d+W)?(\\d+D)?(T(?=\\d+[HM])(\\d+H)?(\\d+M)?)?$",
    )



import logging
import json
from pathlib import Path
from typing import Optional, Union, Any, List

from jsonschema import validate, ValidationError
from whylabs_client.api.notification_settings_api import NotificationSettingsApi
from whylabs_client.api.models_api import ModelsApi

from whylabs_toolkit.monitor.manager.monitor_setup import MonitorSetup
from whylabs_toolkit.monitor.models import *
from whylabs_toolkit.helpers.monitor_helpers import get_model_granularity
from whylabs_toolkit.helpers.config import Config
from whylabs_toolkit.helpers.utils import get_monitor_api, get_notification_api


logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)


class MonitorManager:
    def __init__(
        self,
        setup: MonitorSetup,
        eager: Optional[bool] = None,
        notifications_api: Optional[NotificationSettingsApi] = None,
        monitor_api: Optional[ModelsApi] = None,
        config: Config = Config(),
    ) -> None:
        self._setup = setup
        self.__notifications_api = notifications_api or get_notification_api(config=config)
        self.__monitor_api = monitor_api or get_monitor_api(config=config)
        self.__eager = eager

    def _get_existing_notification_actions(self) -> List[str]:
        actions_dict_list = self.__notifications_api.list_notification_actions(org_id=self._setup.credentials.org_id)
        action_ids = []
        for action in actions_dict_list:
            action_ids.append(action.get("id"))
        return action_ids

    @staticmethod
    def get_notification_request_payload(action: Union[SlackWebhook, EmailRecipient, PagerDuty]) -> str:
        if isinstance(action, SlackWebhook):
            return "slackWebhook"
        elif isinstance(action, EmailRecipient):
            return "email"
        elif isinstance(action, PagerDuty):
            return "pagerDutyKey"
        else:
            raise ValueError(
                f"Can't work with {action} type. Available options are SlackWebhook, PagerDuty and EmailRecipient."
            )

    def _update_notification_actions(self) -> None:
        """
        Updates the notification actions to be passed to WhyLabs based on the actions defined in the MonitorBuilder object.
        """
        if not self._setup.monitor:
            raise ValueError("You must call apply() on your MonitorSetup object!")

        existing_actions = self._get_existing_notification_actions()

        for action in self._setup.monitor.actions:
            if isinstance(action, GlobalAction):
                continue

            if action.id not in existing_actions:
                logger.info(f"Didn't find a {action.type} action under the ID {action.id}, creating one now!")
                payload_key = self.get_notification_request_payload(action=action)
                self.__notifications_api.put_notification_action(
                    org_id=self._setup.credentials.org_id,
                    type=action.type.upper(),
                    action_id=action.id,
                    body={payload_key: action.destination},
                )

        if self._setup.monitor:
            self._setup.monitor.actions = [
                action if isinstance(action, GlobalAction) else GlobalAction(target=action.id)
                for action in self._setup.monitor.actions
            ]

    def _get_current_monitor_config(self) -> Optional[Any]:
        monitor_config = self.__monitor_api.get_monitor_config_v3(
            org_id=self._setup.credentials.org_id, dataset_id=self._setup.credentials.dataset_id
        )
        return monitor_config

    def dump(self) -> Any:
        self._update_notification_actions()

        doc = Document(
            orgId=self._setup.credentials.org_id,
            datasetId=self._setup.credentials.dataset_id,
            granularity=get_model_granularity(
                org_id=self._setup.credentials.org_id, dataset_id=self._setup.credentials.dataset_id  # type: ignore
            ),
            analyzers=[self._setup.analyzer],
            monitors=[self._setup.monitor],
            allowPartialTargetBatches=self.__eager,
        )
        return doc.json(indent=2, exclude_none=True)

    def validate(self) -> bool:
        try:
            Monitor.validate(self._setup.monitor)
            Analyzer.validate(self._setup.analyzer)

            with open(f"{Path(__file__).parent.parent.resolve()}/schema/schema.json", "r") as f:
                schema = json.load(f)
            document = self.dump()
            validate(instance=json.loads(document), schema=schema)
            return True
        except ValidationError as e:
            raise e

    def save(self) -> None:
        if self.validate() is True:
            self.__monitor_api.put_analyzer(
                org_id=self._setup.credentials.org_id,
                dataset_id=self._setup.credentials.dataset_id,
                analyzer_id=self._setup.credentials.analyzer_id,
                body=self._setup.analyzer.dict(exclude_none=True),  # type: ignore
            )
            self.__monitor_api.put_monitor(
                org_id=self._setup.credentials.org_id,
                dataset_id=self._setup.credentials.dataset_id,
                monitor_id=self._setup.credentials.monitor_id,
                body=self._setup.monitor.dict(exclude_none=True),  # type: ignore
            )
        if self.__eager is not None:
            current_config = self._get_current_monitor_config()

            if self.__eager != current_config.get("allowPartialTargetBatches"):  # type: ignore
                current_config["allowPartialTargetBatches"] = self.__eager  # type: ignore
                self.__monitor_api.put_monitor_config_v3(
                    org_id=self._setup.credentials.org_id,
                    dataset_id=self._setup.credentials.dataset_id,
                    body=current_config,
                )



# flake8: noqa
"""Console script for monitor_schema."""
from .analyzer import *
from .column_schema import *
from .commons import *
from .document import *
from .monitor import *
from .segments import *
from ...utils.granularity import Granularity

# TODO add all algorithms


__all__ = [
    "DatasetMetric",
    "SimpleColumnMetric",
    "ComplexMetrics",
    "Metadata",
    # analyzer
    "Analyzer",
    # baseline
    "BaselineType",
    "ReferenceProfileId",
    "TimeRangeBaseline",
    "TimeRange",
    "TrailingWindowBaseline",
    "SingleBatchBaseline",
    # configs
    "DiffConfig",
    "DriftConfig",
    "ComparisonConfig",
    "ComparisonOperator",
    "FrequentStringComparisonConfig",
    "FrequentStringComparisonOperator",
    "ListComparisonOperator",
    "ListComparisonConfig",
    "ExperimentalConfig",
    "FixedThresholdsConfig",
    "ColumnListChangeConfig",
    "SeasonalConfig",
    "StddevConfig",
    # targets
    "DatasetMatrix",
    "ColumnMatrix",
    "Segment",
    "SegmentTag",
    "TargetLevel",
    # monitors
    "Monitor",
    "EveryAnomalyMode",
    "DigestMode",
    "AnomalyFilter",
    # scheduling
    "ImmediateSchedule",
    "CronSchedule",
    "FixedCadenceSchedule",
    "Cadence",
    # monitor actions
    # "RawWebhook",
    "SlackWebhook",
    "EmailRecipient",
    "PagerDuty",
    "GlobalAction",
    # big document
    "Document",
    # schema
    "EntitySchema",
    "ColumnSchema",
    "ColumnDataType",
    "ColumnDiscreteness",
    "WeightConfig",
    "SegmentWeightConfig",
    "Granularity",
    # enums
    "DiffMode",
    "ThresholdType",
    "AlgorithmType",
    "DatasetMetric",
    "SimpleColumnMetric",
    "ComplexMetrics",
    "ExpectedValue",
]



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['whylabs_toolkit',
 'whylabs_toolkit.cli',
 'whylabs_toolkit.container',
 'whylabs_toolkit.helpers',
 'whylabs_toolkit.monitor',
 'whylabs_toolkit.monitor.manager',
 'whylabs_toolkit.monitor.models',
 'whylabs_toolkit.monitor.models.analyzer',
 'whylabs_toolkit.utils']

package_data = \
{'': ['*'], 'whylabs_toolkit.monitor': ['schema/*']}

install_requires = \
['jsonschema>=4.17.3,<5.0.0',
 'pydantic>=1.10.4,<2.0.0',
 'whylabs-client>=0.5,<0.6',
 'whylogs>=1.1.26,<2.0.0']

setup_kwargs = {
    'name': 'whylabs-toolkit',
    'version': '0.0.14',
    'description': 'Whylabs CLI and Helpers package.',
    'long_description': "# WhyLabs Toolkit\n\nThe WhyLabs Toolkit package contains helper methods to help users interact with our internal APIs. Users will benefit from using it if they want to abstract some of WhyLabs' internal logic and also automate recurring API calls.\n\n\n## Basic usage\nTo start using the `whylabs_toolkit` package, install it from PyPI with:\n```bash\npip install whylabs_toolkit\n``` \n\n## Packages\n\nThe available packages that we have enable different use-cases for the `whylabs_toolkit`. To get started, navigate to one of the following sections and find useful tutorials there.\n\n| Package             | Usage                |\n|---------------------|----------------------|\n| [Monitor Manager](https://github.com/whylabs/whylabs-toolkit/blob/mainline/whylabs_toolkit/monitor/manager/README.md) | Author and modify existing WhyLabs monitor with Python |\n| [WhyLabs Helpers](https://github.com/whylabs/whylabs-toolkit/blob/mainline/whylabs_toolkit/helpers/README.md) | Interact with and modify your Datasets and ML Models specs in WhyLabs. |\n\n## Development\n\nTo start contributing, you will manage dependencies with [Poetry](https://python-poetry.org/) and also a handful of `Makefile` commands. To install all necessary dependencies and activate the virtual environment, run:\n\n```bash\nmake setup && poetry shell\n```\n\n## Get in touch\nIf you want to learn more how you can benefit from this package or if there is anything missing, please [contact our support](https://whylabs.ai/contact-us), we'll be more than happy to help you!",
    'author': 'Anthony Naddeo',
    'author_email': 'anthony.naddeo@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



import logging
from datetime import datetime
from typing import Optional, Union

from whylabs_client.api.dataset_profile_api import DeleteDatasetProfilesResponse, DeleteAnalyzerResultsResponse

from whylabs_toolkit.helpers.utils import get_dataset_profile_api
from whylabs_toolkit.helpers.config import Config

date_or_millis = Union[datetime, int]
logger = logging.getLogger(__name__)


def validate_timestamp_in_millis(epoch_milliseconds: int) -> bool:
    if not isinstance(epoch_milliseconds, int):
        return False
    try:
        epoch_seconds = epoch_milliseconds / 1000
        dt = datetime.fromtimestamp(epoch_seconds)
        return dt >= datetime(1970, 1, 1)
    except (ValueError, OverflowError):
        return False


def process_date_input(date_input: date_or_millis) -> int:
    if isinstance(date_input, int):
        try:
            assert validate_timestamp_in_millis(epoch_milliseconds=date_input)
            return date_input
        except AssertionError:
            raise ValueError("You must provide a valid date input")
    elif isinstance(date_input, datetime):
        return int(date_input.timestamp() * 1000.0)
    else:
        raise ValueError(f"The date object {date_input} input must be a datetime or an integer Epoch!")


def delete_all_profiles_for_period(
    start: date_or_millis,
    end: date_or_millis,
    config: Config = Config(),
    org_id: Optional[str] = None,
    dataset_id: Optional[str] = None,
) -> DeleteDatasetProfilesResponse:
    api = get_dataset_profile_api()

    profile_start_timestamp = process_date_input(date_input=start)
    profile_end_timestamp = process_date_input(date_input=end)

    org_id = org_id or config.get_default_org_id()
    dataset_id = dataset_id or config.get_default_dataset_id()

    result_profiles: DeleteDatasetProfilesResponse = api.delete_dataset_profiles(
        org_id=org_id,
        dataset_id=dataset_id,
        profile_start_timestamp=profile_start_timestamp,
        profile_end_timestamp=profile_end_timestamp,
    )
    logger.info(f"Scheduled deletion for profiles on {dataset_id} for {org_id}")

    api.delete_analyzer_results(
        org_id=org_id,
        dataset_id=dataset_id,
        start_timestamp=profile_start_timestamp,
        end_timestamp=profile_end_timestamp,
    )

    logger.info("Deleted analyzer results for the same timestamps as the profiles")
    logger.info(f"NOTE: Profile deletion happens every full hour on WhyLabs")

    return result_profiles



from typing import Optional
from dataclasses import dataclass, field

from whylabs_toolkit.helpers.config import Config


@dataclass
class MonitorCredentials:
    monitor_id: str
    dataset_id: Optional[str] = field(default=None)  # type: ignore
    config: Config = field(default=Config())  # type: ignore

    def __post_init__(self) -> None:
        self.org_id = self.config.get_default_org_id()
        self.analyzer_id = f"{self.monitor_id}-analyzer"
        if not self.dataset_id:
            self.dataset_id = self.config.get_default_dataset_id()



from whylabs_client import ApiClient, Configuration

from .config import Config


def create_client(config: Config = Config()) -> ApiClient:
    client_config = Configuration(host=config.get_whylabs_host())
    client_config.api_key = {"ApiKeyAuth": config.get_whylabs_api_key()}
    client_config.discard_unknown_keys = True
    return ApiClient(client_config)



import os
import logging
from enum import Enum

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)


class ConfigVars(Enum):
    ORG_ID = 1
    DATASET_ID = 2
    WHYLABS_API_KEY = 3
    WHYLABS_HOST = "https://api.whylabsapp.com"
    WHYLABS_PRIVATE_API_ENDPOINT = 5


class Config:
    def get_whylabs_api_key(self) -> str:
        return Validations.require(ConfigVars.WHYLABS_API_KEY)

    def get_whylabs_host(self) -> str:
        _private_api_endpoint = Validations.get_or_default(ConfigVars.WHYLABS_PRIVATE_API_ENDPOINT)
        if _private_api_endpoint and isinstance(_private_api_endpoint, str):
            logger.debug(f"Using private API endpoint: {_private_api_endpoint}")
            return _private_api_endpoint
        return Validations.get_or_default(ConfigVars.WHYLABS_HOST)

    def get_default_org_id(self) -> str:
        return Validations.require(ConfigVars.ORG_ID)

    def get_default_dataset_id(self) -> str:
        return Validations.require(ConfigVars.DATASET_ID)


class UserConfig(Config):
    def __init__(self, api_key: str, org_id: str, dataset_id: str, whylabs_host: str = ConfigVars.WHYLABS_HOST.value):
        self.api_key = api_key
        self.whylabs_host = whylabs_host
        self.org_id = org_id
        self.dataset_id = dataset_id

    def get_whylabs_api_key(self) -> str:
        return self.api_key

    def get_whylabs_host(self) -> str:
        return self.whylabs_host

    def get_default_org_id(self) -> str:
        return self.org_id

    def get_default_dataset_id(self) -> str:
        return self.dataset_id


class Validations:
    @staticmethod
    def require(env: ConfigVars) -> str:
        val = os.getenv(env.name)
        if not val:
            raise TypeError(f"Missing {env.name} env variable.")
        return val

    @staticmethod
    def get_or_default(env: ConfigVars) -> str:
        val = os.getenv(env.name, env.value)
        if not val:
            raise TypeError(f"No default value for {env.name}")
        return val
